/* 
 *  send signal n (by sending n*USR2 + USR1) to the running process
 *  define my_getpid() to identify threads
 */

#include <stdlib.h>
#include <stdio.h>

/* next two includes needed for getpid(), outcommend on problems */
#include <sys/types.h>
#include <unistd.h>

/* outcomment next include, if your system does not support it */
#ifndef NoSIGNAL     /* may be defined by configure in Makefile */
#include <signal.h>  /* broken for SLES9 g++-4.1 at IA-64 */
/*  linux: _SIGNAL_H, OSF1: _SYS_SIGNAL_H_ */
#endif

/* simply outcomment next line, if you have problems with time.h */
#include <time.h>	/* defines time() and clock() */
/* time.h often defines ulong (therefore we use ULong) */

#include "error.h"
#include "sighand.h"

/* export: int sig1; int ini_handler(); */
volatile int sig1;	/* set by signal handler */

unsigned long starttime=0l;
#if defined(_TIME_H) || defined(_TIME_H_)
/* for the case that time.h was not included, skip time code
 *  linux: _TIME_H, OSF1: _TIME_H_,_SYS_SIGNAL_H_
 */
long itime(){	/* unixzeit=time */
  time_t i; time(&i); /* multithread safe? ToDo 2013: seconds as double + us */
  if (starttime==0l) starttime=i;
  return(long)(i-starttime);
}
#else
#warning "no time.h, time code is dropped"
long itime(){ return 0; }
#endif

/* 2013-04 add more precise time for performance measurements */
#if     defined(CONFIG_MPI)
#include <mpi.h>
double fstime(){ return MPI_Wtime(); } /* seconds as double */
#elif   1
#include <sys/time.h>
double fstime(){
 struct timeval tv; if(gettimeofday(&tv, NULL))F0("fstime");
 return tv.tv_sec+1e-6*tv.tv_usec; }
#else
#warning no time measurements
double fstime(){ return 0.0; } /* dummy 0.0 */
#endif

/*
+++TeX+++
 This part of the program allows the user to control
 the running process a bit. You are able to stop some
 functions in a controled manner or get further information,
 what the running process is doing.
	
 This is done by using unix signals (SIGUSR1 and SIGUSR2).
 On unix systems you can do it via the kill command from the command line,
 which are usually used to stop or continue a unix process.
	
 LANG=de (have to translate it):
 
	Ist die Prozess-Identifikations-Nummer (PID) zum Beispiel 1234, so
	kann mit {\bf kill -SIGUSR1 1234} oder {\bf kill -SIGUSR2 1234}
	eines der beiden Signale an den Prozess gesendet werden.
	
	Mit einer Kombination von $n$ Signalen SIGUSR2 und einem
	SIGUSR1 koennen auch kompliziertere Reaktionen vom Programm 
	erzeugt werden, da das Programm die Variable {\bf sig1}, die
	dann $n$ enthaelt, jederzeit auslesen kann.
---TeX---
*/

#define PID_FILE "spins.pid"

/* this is a frame for calling getpid(), can be used to identify thread?
 * all sigs work only on mainPID (linux, ini_handler before/after pthread create)
 *  (ToDo: ini_handler must be called by each thread?)
 */
long my_getpid() {
#if defined(_UNISTD_H) || defined(_UNISTD_H_)
  return (long)getpid();
#else
#warning "no unistd.h, getpid code was dropped"
  return 0; /* not supported */
#endif
}

extern int mpi_i; /* MPI = Message Passing Interface, mpirun */
#if defined(_SIGNAL_H) || defined(_SIGNAL_H_) || defined(_SYS_SIGNAL_H_)
void my_handler(int sig){ /* changed behav. 2015-02-26 */
  switch(sig){
    case SIGUSR1: sig1+=0x0010; break; // conflicts to blcr (checkpointing)
    case SIGUSR2: sig1+=0x0001; break;
    case SIGALRM: sig1+=0x0100;/* 256 = alarm(60); */ break; 
    default: sig1+=0x1000; /* 4096 */
  }
  if ((mpi_i&(mpi_i-1))==0) // log2 threads do reporting, 2016-10 moved
    printf("\nDBG PID= %5ld MPI= %2d got signal %2d, sig1= %04x ",
      my_getpid(), mpi_i, sig, sig1);
  signal(sig,my_handler); /* redirect again */}
#endif

/* multiple call possible */
/* ToDo: save mainPID and if PID != mainPID I am a child thread */
int ini_handler(int mpi_i, int vvv){ /* better open and close_handler ??? */
   FILE *f1; long mypid=my_getpid();
   /* store the own PID (Process IDentification number) in a file */
   /* can also be used, too communicate with the process (sh status) */
   if (!mpi_i) { /* only node 0 for simplicity */
     f1=fopen(PID_FILE,"w");
     if (!f1) fprintf(stderr," error ini_handler: creating " PID_FILE );
     if (f1) { if (vvv) printf(" write " PID_FILE "\n");
       /* in error case mypid is 0, user cant sent a signal to process 0 */
       fprintf(f1,"%ld", mypid );
       fclose(f1); }
    }
   sig1=0;
#if defined(_SIGNAL_H) || defined(_SIGNAL_H_) || defined(_SYS_SIGNAL_H_)
//   signal(SIGUSR1,my_handler);	/* sig1+=0x0001; */ // blcr conflicts
   signal(SIGUSR2,my_handler);  /* sig1+=0x0010; */
   signal(SIGALRM,my_handler);  /* sig1+=0x0100; */
   /* if(vvv&16)alarm(60); */	/* Signal SIGALRM in 60s */
#else
#warning "no signal.h, communication code dropped"
#endif
   return 0;
}
